<?php
/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Entities;

use JsonSerializable;

/**
 * IQRF OS and DPA entity
 */
final class OsDpa implements JsonSerializable {

	/**
	 * @var string IQRF OS build
	 */
	private $os;

	/**
	 * @var string DPA version
	 */
	private $dpa;

	/**
	 * @var string Notes
	 */
	private $notes;

	/**
	 * @var string Download path
	 */
	private $downloadPath;

	/**
	 * Constructor
	 * @param string $os IQRF OS build
	 * @param string $dpa DPA version
	 * @param string $notes Notes
	 * @param string $downloadPath Download path
	 */
	public function __construct(string $os, string $dpa, string $notes, string $downloadPath) {
		$this->os = $os;
		$this->dpa = $dpa;
		$this->notes = $notes;
		$this->downloadPath = $downloadPath;
	}

	/**
	 * Creates a new IQRF OS and DPA entity from an array
	 * @param array<string,mixed> $data JSON data in an array
	 * @return OsDpa Created IQRF OS and DPA entity
	 */
	public static function fromArray(array $data): self {
		return new static($data['os'], $data['dpa'], $data['notes'], $data['downloadPath']);
	}

	/**
	 * Returns the IQRF OS build
	 * @return string IQRF OS build
	 */
	public function getOs(): string {
		return $this->os;
	}

	/**
	 * Returns the DPA version
	 * @param bool $pretty Pretty formatting
	 * @return string DPA version
	 */
	public function getDpa(bool $pretty = false): string {
		if ($pretty) {
			$versionInt = hexdec($this->dpa);
			$major = strval($versionInt >> 8);
			$minor = str_pad(dechex($versionInt & 0xff), 2, '0', STR_PAD_LEFT);
			return $major . '.' . $minor;
		}
		return $this->dpa;
	}

	/**
	 * Returns the notes
	 * @return string Notes
	 */
	public function getNotes(): string {
		return $this->notes;
	}

	/**
	 * Returns the download path
	 * @return string Download path
	 */
	public function getDownloadPath(): string {
		return $this->downloadPath;
	}

	/**
	 * Serialized the IQRF OS and DPA entity into JSON
	 * @return array<string, string> JSON serialized IQRF OS and DPA entity
	 */
	public function jsonSerialize(): array {
		return [
			'osBuild' => $this->os,
			'dpa' => $this->getDpa(true),
			'dpaRaw' => $this->dpa,
			'notes' => $this->notes,
			'downloadPath' => $this->downloadPath,
		];
	}

}
