<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Entities;

use DateTime;
use Iqrf\Repository\Exceptions\CorruptedData;
use Throwable;

/**
 * Server status entity
 */
final class ServerStatus {

	/**
	 * @var int API version
	 */
	private $apiVersion;

	/**
	 * @var string Server hostname
	 */
	private $hostname;

	/**
	 * @var string Authenticated user name
	 */
	private $user;

	/**
	 * @var DateTime Server build date and time
	 */
	private $buildDateTime;

	/**
	 * @var DateTime Server startup date and time
	 */
	private $startDateTime;

	/**
	 * @var DateTime Current date and time
	 */
	private $dateTime;

	/**
	 * @var int All database tables checksum
	 */
	private $databaseChecksum;

	/**
	 * @var DateTime Last database change date and time
	 */
	private $databaseChangeDateTime;

	/**
	 * Constructor
	 * @param int $apiVersion API version
	 * @param string $hostname Server hostname
	 * @param string $user Authenticated user name
	 * @param string $buildDateTime Server build date and time
	 * @param string $startDateTime Server startup date and time
	 * @param string $dateTime Current date and time
	 * @param int $databaseChecksum All database tables checksum
	 * @param string $databaseChangeDateTime Last database change date and time
	 * @throws CorruptedData
	 */
	public function __construct(int $apiVersion, string $hostname, string $user, string $buildDateTime, string $startDateTime, string $dateTime, int $databaseChecksum, string $databaseChangeDateTime) {
		$this->apiVersion = $apiVersion;
		$this->hostname = $hostname;
		$this->user = $user;
		$this->databaseChecksum = $databaseChecksum;
		try {
			$this->buildDateTime = new DateTime($buildDateTime);
			$this->startDateTime = new DateTime($startDateTime);
			$this->dateTime = new DateTime($dateTime);
			$this->databaseChangeDateTime = new DateTime($databaseChangeDateTime);
		} catch (Throwable $e) {
			throw new CorruptedData();
		}
	}

	/**
	 * Creates a new server status entity from an array
	 * @param array<string,mixed> $data JSON data in an array
	 * @return ServerStatus Created server status entity
	 * @throws CorruptedData
	 */
	public static function fromArray(array $data): self {
		return new static($data['apiVersion'], $data['hostname'], $data['user'], $data['buildDateTime'], $data['startDateTime'], $data['dateTime'], $data['databaseChecksum'], $data['databaseChangeDateTime']);
	}

	/**
	 * Returns the API version
	 * @return int API version
	 */
	public function getApiVersion(): int {
		return $this->apiVersion;
	}

	/**
	 * Returns the server hostname
	 * @return string Server hostname
	 */
	public function getHostname(): string {
		return $this->hostname;
	}

	/**
	 * Returns the authenticated user name
	 * @return string Authenticated user name
	 */
	public function getUser(): string {
		return $this->user;
	}

	/**
	 * Returns the server build date and time
	 * @return DateTime Server build date and time
	 */
	public function getBuildDateTime(): DateTime {
		return $this->buildDateTime;
	}

	/**
	 * Returns the server startup date and time
	 * @return DateTime Server startup date and time
	 */
	public function getStartDateTime(): DateTime {
		return $this->startDateTime;
	}

	/**
	 * Returns the current date and time
	 * @return DateTime Current date and time
	 */
	public function getDateTime(): DateTime {
		return $this->dateTime;
	}

	/**
	 * Returns the all database tables checksum
	 * @return int All database tables checksum
	 */
	public function getDatabaseChecksum(): int {
		return $this->databaseChecksum;
	}

	/**
	 * Returns the last database change date and time
	 * @return DateTime Last database change date and time
	 */
	public function getDatabaseChangeDateTime(): DateTime {
		return $this->databaseChangeDateTime;
	}

}
