<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Models;

use Iqrf\Repository\Entities\OsDpa;
use Iqrf\Repository\Exceptions\CorruptedData;
use Iqrf\Repository\Exceptions\ServiceUnavailable;

/**
 * IQRF OS and DPA versions manager
 */
final class OsAndDpaManager extends BaseManager {

	/**
	 * IQRF OS and DPA versions path
	 */
	private const PATH = 'osdpa';

	/**
	 * Lists all IQRF OS and DPA versions
	 * @return array<OsDpa> IQRF OS and DPA versions
	 * @throws CorruptedData
	 * @throws ServiceUnavailable
	 */
	public function list(): array {
		$response = $this->apiClient->sendRequest(self::PATH);
		return $this->parseResponse($response);
	}

	/**
	 * Parses the JSON API response
	 * @param array<mixed> $response JSON API response
	 * @return array<OsDpa> IQRF OS and DPA versions
	 */
	private function parseResponse(array $response): array {
		foreach ($response as &$item) {
			$item = OsDpa::fromArray($item);
		}
		return $response;
	}

	/**
	 * Returns filtered IQRF OS and DPA versions
	 * @param string|null $os IQRF OS build
	 * @param string|null $dpa DPA versions
	 * @return array<OsDpa> IQRF OS and DPA versions
	 * @throws CorruptedData
	 * @throws ServiceUnavailable
	 */
	public function get(?string $os = null, ?string $dpa = null): array {
		$query = [];
		if ($os !== null) {
			$query['os'] = $os;
		}
		if ($dpa !== null) {
			$query['dpa'] = $dpa;
		}
		$response = $this->apiClient->sendRequest(self::PATH, $query);
		return $this->parseResponse($response);
	}

}
