<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Models;

use BadMethodCallException;
use GuzzleHttp\Client;
use Iqrf\Repository\Entities\File;
use Iqrf\Repository\Entities\Files;
use Nette\Utils\Json;
use Nette\Utils\JsonException;

/**
 * Files manager
 */
final class FilesManager {

	/**
	 * @var Client HTTP(S) client
	 */
	private $httpClient;

	/**
	 * @var string|null Base path
	 */
	private $path = null;

	/**
	 * Constructor
	 * @param Client $httpClient HTTP(S) client
	 */
	public function __construct(Client $httpClient) {
		$this->httpClient = $httpClient;
	}

	/**
	 * Returns the base path
	 * @return string|null Base path
	 */
	public function getPath(): ?string {
		return $this->path;
	}

	/**
	 * Sets the base path
	 * @param string $path Base path
	 */
	public function setPath(string $path): void {
		$this->path = $path;
	}

	/**
	 * Download the file and returns its content
	 * @param string $fileName File name
	 * @return string|null File content
	 * @throws JsonException
	 */
	public function download(string $fileName): ?string {
		$file = $this->get($fileName);
		if ($file === null) {
			return null;
		}
		$url = $this->path . '/' . $file->getName();
		return $this->httpClient->get($url)->getBody()->getContents();
	}

	/**
	 * Return the file entity
	 * @param string $fileName File name
	 * @return File|null File
	 * @throws JsonException
	 */
	public function get(string $fileName): ?File {
		$files = $this->list();
		$name = ltrim($fileName, './');
		foreach ($files->getFiles() as $file) {
			if ($file->getName() === $name) {
				return $file;
			}
		}
		return null;
	}

	/**
	 * Lists files
	 * @return Files Files
	 * @throws JsonException
	 */
	public function list(): Files {
		if ($this->path === null) {
			throw new BadMethodCallException();
		}
		$url = $this->path . '/__files__.json';
		$json = $this->httpClient->get($url)->getBody()->getContents();
		$array = Json::decode($json, Json::FORCE_ARRAY);
		return Files::fromArray($array);
	}

}
